import magicgui
import napari
import napari.settings
from qtpy.QtCore import QEventLoop


class InteractiveUI:
    def __init__(self):
        self.viewer = napari.Viewer()
        self.event_loop = QEventLoop()
        self.widget = None

    def continue_event(self, last):
        self.viewer.window.remove_dock_widget(self.widget)
        if last:
            self.viewer.close()

        self.event_loop.quit()

    def run_widget(self, func, auto_call=False, last=False):
        # Make the widget.
        widget_func = magicgui.magicgui(func, auto_call=auto_call)
        self.widget = self.viewer.window.add_dock_widget(widget_func, area="right")
        # Setup the continue button.
        continue_btn = magicgui.widgets.PushButton(value=True, text="Continue")
        continue_btn.changed.connect(lambda: self.continue_event(last=last))
        widget_func.append(continue_btn)
        # Call the widget function to initialize it in the viewer.
        widget_func()

        # If we are in a notebook we don't want napari to integrate with the
        # ipython loop since we want it to run in qt.
        settings = napari.settings.get_settings().application
        ipy = settings.ipy_interactive
        settings.ipy_interactive = False
        self.viewer.show()
        # Reset ipy_interactive to its previous value.
        settings.ipy_interactive = ipy
        # Run the event loop.
        self.event_loop.exec_()

        # Getting the function's return value re-adds the layers so we need to clear them after.
        retval = widget_func()
        self.viewer.layers.clear()
        return retval



import xarray as xr

import magnify.registry as registry


@registry.component("drop")
def drop(
    xp: xr.Dataset,
    roi_only: bool = False,
    drop_tiles: bool = True,
):
    if roi_only:
        return xp.roi.assign_attrs(xp.attrs)
    elif drop_tiles:
        return xp.drop_vars(["tile", "tile_row", "tile_col"], errors="ignore")
    else:
        return xp


@registry.component("restore_format")
def restore_format(xp: xr.Dataset | xr.DataArray) -> xr.Dataset | xr.DataArray:
    xp = xp.unstack()

    if "__time__" in xp.dims:
        # Restore the time coordinate to its original name if it got changed to not clash.
        xp = xp.rename({"__time__": "time"})

    # Remove any dimensions that got added by standardize_format.
    standard_dims = ["channel", "time", "tile_row", "tile_col", "tile_y", "tile_x"]
    for dim in standard_dims:
        if dim not in xp.__original_tile_dims__ and dim in xp.dims:
            xp = xp.squeeze(dim)

    # Restore the original dimension order in the context of dimensions that didn't exist before.
    for name, var in xp.variables.items():
        original_dims = [dim for dim in xp.__original_tile_dims__ if dim in var.dims]
        # If there are no original dimensions in this variable then don't do anything.
        if len(original_dims) > 0:
            var_dims = list(var.dims)
            # We guarantee that original dimensions are contiguous so find their start and end.
            idxs = [i for i, x in enumerate(var_dims) if x in original_dims]
            start, end = (idxs[0], idxs[-1] + 1)
            # Keep all non-original dimensions in the same order while permuting the originals.
            dim_order = var_dims[:start] + original_dims + var_dims[end:]
            xp[name] = var.transpose(*dim_order)

    del xp.attrs["__original_tile_dims__"]

    return xp



from __future__ import annotations

import re

import numba
import numpy as np
import pandas as pd
import scipy

import magnify.registry as registry


@registry.component("identify_buttons")
def identify_buttons(assay, shape=None, pinlist=None, blank=None):
    if blank is None:
        blank = ["", "blank", "BLANK"]

    if pinlist is not None:
        df = pd.read_csv(pinlist)
        df["Indices"] = df["Indices"].apply(
            lambda s: [int(x) for x in re.sub(r"[\(\)]", "", s).split(",")]
        )
        # Replace blanks with the empty string.
        df["MutantID"] = df["MutantID"].replace(blank, "")
        # Zero-index the indices.
        cols, rows = np.array(df["Indices"].to_list()).T - 1
        names = df["MutantID"].to_numpy(dtype=str, na_value="")
        names_array = np.empty((max(rows) + 1, max(cols) + 1), dtype=names.dtype)
        names_array[rows, cols] = names
    elif shape is not None:
        names_array = np.empty((shape[0], shape[1]), dtype="<U200")
        names_array.fill("default")
    else:
        raise ValueError("Either pinlist or shape must be provided.")

    assay = assay.assign_coords(
        tag=(("mark_row", "mark_col"), names_array),
        valid=(
            ("mark_row", "mark_col", "time"),
            np.ones(
                (names_array.shape[0], names_array.shape[1], assay.sizes["time"]),
                dtype=bool,
            ),
        ),
    )

    return assay


@registry.component("identify_mrbles")
def identify_mrbles(assay, spectra, codes, reference="eu"):
    # Read in the dataframe of lanthanide spectra and make sure the reference lanthanide is first.
    spectra_df = pd.read_csv(spectra)
    ref_matches = spectra_df[spectra_df["name"] == reference].index
    if len(ref_matches) == 0:
        raise ValueError(f"Reference lanthanide '{reference}' not found in spectra file")
    ref_idx = ref_matches[0]
    spectra_df = spectra_df.reindex([ref_idx] + [i for i in range(len(spectra_df)) if i != ref_idx])
    lns = spectra_df["name"].to_list()
    num_lns = len(lns)

    # Read in the dataframe of codes.
    codes_df = pd.read_csv(codes)
    tag_names = codes_df["name"].to_numpy()
    num_codes = len(tag_names)
    # Make sure spectra and codes have consistent lanthanides.
    code_lns = set(codes_df.columns)
    code_lns.remove("name")
    if code_lns != set(lns):
        raise ValueError(f"Lanthanide names in {codes} do not match lanthanide names in {spectra}.")

    # Step 1: Estimate the lanthanide volumes in each bead by solving the linear equation SV = I
    # where S are the reference spectra and I are the intensities of each bead.
    channels = [c for c in assay.channel.values if c in spectra_df.columns]
    sp = spectra_df[channels].to_numpy()
    sel = assay.roi.isel(time=0).sel(channel=channels)
    intensities = (
        sel.where(sel.fg).mean(dim=["roi_x", "roi_y"])
        - sel.where(sel.bg).median(dim=["roi_x", "roi_y"])
    ).to_numpy()
    volumes = np.linalg.lstsq(sp.T, intensities.T, rcond=None)[0].T
    # We also want the lanthanide ratios with respect to the reference lanthanide.
    ratios = volumes / volumes[:, 0:1]
    # Save the resulting lanthanide volumes and ratios to the assay.
    assay = assay.assign_coords(ln=("ln", lns))
    assay["ln_vol"] = (("mark", "ln"), volumes)
    assay["ln_ratio"] = (("mark", "ln"), ratios)

    # Step 2: Agressively remove outliers to make future processing easier.
    X = ratios[:, 1:]
    # Find the distance to a point that should still be in the same cluster assuming cluster
    # sizes differ by a factor of at most 20 from the mean cluster size.
    n_neighbor = round(len(X) / (20 * num_codes)) + 2
    dist = (
        scipy.spatial.KDTree(X, leafsize=n_neighbor)
        .query(X, k=[n_neighbor], workers=-1)[0]
        .flatten()
    )
    # Treat 5% of points as outliers for this first step.
    X_r = X[dist <= np.percentile(dist, 95)]

    # Step 3: Find an affine transformation of the code's lanthanide ratios to get a clustering
    # that minimizes the distance between each bead and its closest code.
    code_ratios = codes_df[lns[1:]].to_numpy()

    # Minimize the loss only considering affine transforms close to an estimated scaling factor.
    @numba.njit
    def fit_1d(points, codes, counts, N=100):
        if len(codes) == 1:
            return 1, points.mean()

        scale = (points.max() - points.min()) / (codes.max() - codes.min())
        sizes = np.zeros(len(codes))
        dists = np.ones(len(codes))
        best_a = 0
        best_p = 0
        best_cost = np.inf
        for a in np.linspace(0.75 * scale, 1.25 * scale, N):
            for p in np.linspace(points.min(), 0.25 * points.max() + 0.75 * points.min(), N):
                clusters = a * codes + p
                curr_start = 0
                for i in range(len(clusters)):
                    if i < len(clusters) - 1:
                        cluster_mid = (clusters[i] + clusters[i + 1]) / 2
                    else:
                        cluster_mid = np.inf

                    for j in range(curr_start, len(points)):
                        if points[j] > cluster_mid:
                            break

                    if curr_start == j:
                        dists[i] = np.inf
                    else:
                        dists[i] = ((points[curr_start : j + 1] - clusters[i]) ** 2).mean()
                    sizes[i] = j - curr_start
                    curr_start = j

                cost = (
                    100 * dists.mean() + ((sizes / sizes.sum() - counts / counts.sum()) ** 2).mean()
                )
                if cost < best_cost:
                    best_a = a
                    best_p = p
                    best_cost = cost

        return best_a, best_p

    A = np.zeros(num_lns - 1)
    p = np.zeros(num_lns - 1)
    for i in range(num_lns - 1):
        c, counts = np.unique(code_ratios[:, i], return_counts=True)
        A[i], p[i] = fit_1d(np.sort(X_r[:, i]), c, counts)

    # Cluster points to the closest code.
    tag_idxs = np.argmin(
        np.linalg.norm(X_r[:, np.newaxis] - (A * code_ratios + p)[np.newaxis], axis=-1),
        axis=1,
    )

    # Step 4: Perform a better clustering using a Gaussian mixture model initialized with the
    # clustering from step 3. We also add a uniform distribution to the mixture which allows us to
    # exclude outliers less agressively.
    num_codes = len(code_ratios)
    means = np.zeros((num_codes, num_lns - 1))
    covs = np.zeros((num_codes, num_lns - 1, num_lns - 1)) + np.eye(num_lns - 1) * 1e-10
    proportions = np.zeros(num_codes + 1)

    # Initialize the Gaussian components.
    for i in range(num_codes):
        proportions[i] = np.sum(tag_idxs == i) + 1
        means[i] = np.median(X_r[tag_idxs == i], axis=0)
        if proportions[i] > 1:
            covs[i] += np.cov(X_r[tag_idxs == i], rowvar=False)

    # Set all component variances to be identical since initializing individual
    # covariances can lead to huge terms.
    covs[:] = np.median(covs, axis=0)
    # Initialize the uniform component.
    proportions[-1] = 1e-10
    proportions /= proportions.sum()
    log_cond_probs = np.empty((len(X), num_codes + 1))
    log_cond_probs[:, -1] = -np.log(X_r.max(axis=0) - X_r.min(axis=0)).sum()
    probs = None

    tag_names = np.append(tag_names, "outlier")
    # Run the Expectation-Maximization algorithm.
    for _ in range(50):
        # E-step: Compute the probability of each point belonging to each component.
        diff = X[:, np.newaxis, :] - means[np.newaxis, :, :]
        # Work in log space most of the time to avoid numerical issues.
        try:
            log_cond_probs[:, :-1] = (
                -X.shape[1] * np.log(2 * np.pi) / 2
                - 0.5 * np.log(np.linalg.det(covs))
                - 0.5 * np.einsum("...i,...ij,...j->...", diff, np.linalg.inv(covs), diff)
            )
        except np.linalg.LinAlgError:
            print("Warning: Code clustering did not converge.")
            break
        log_probs = np.log(proportions) + log_cond_probs
        log_probs -= scipy.special.logsumexp(log_probs, axis=1)[:, np.newaxis]
        probs = np.exp(log_probs)
        # M-step: Update the parameters of each component.
        means = (
            np.sum(probs[:, :-1, np.newaxis] * X[:, np.newaxis, :], axis=0)
            / np.sum(probs[:, :-1], axis=0)[:, np.newaxis]
        )
        diff = X[:, np.newaxis, :] - means[np.newaxis, :, :]
        covs = (
            np.sum(
                probs[:, :-1, np.newaxis, np.newaxis] * np.einsum("...i,...j->...ij", diff, diff),
                axis=0,
            )
            / np.sum(probs[:, :-1], axis=0)[:, np.newaxis, np.newaxis]
        )
        # Regularize the covariance matrix to avoid numerical issues.
        covs += np.eye(X.shape[1]) * np.median(covs) / 10
        proportions = np.sum(probs, axis=0) / X.shape[0]

    # Assign each bead a code based on the clustering we just found.
    tag_names = np.append(tag_names, "outlier")
    if probs is not None:
        tag_idxs = np.argmax(probs, axis=1)
    else:
        tag_idxs = np.argmin(
            np.linalg.norm(X[:, np.newaxis] - (A * code_ratios + p)[np.newaxis], axis=-1),
            axis=1,
        )
    assay = assay.assign_coords(
        tag=("mark", tag_names[tag_idxs]),
    )

    return assay



import functools
import inspect
from io import StringIO
from os import PathLike

import catalogue
import xarray as xr
from numpy.typing import ArrayLike

from magnify.pipeline import Pipeline

readers = catalogue.create("magnify", "readers")
components = catalogue.create("magnify", "components")


def component(name):
    def component_decorator(func):
        @functools.wraps(func)
        def component_factory(*args, **kwargs):
            return functools.partial(func, *args, **kwargs)

        # Make the factory signature identical to func's except for the first argument.
        signature = inspect.signature(func)
        signature = signature.replace(parameters=list(signature.parameters.values())[1:])
        component_factory.__signature__ = signature
        components.register(name)(component_factory)
        return func

    return component_decorator


def microfluidic_chip(
    data: ArrayLike | str,
    shape: tuple[int, int] = (8, 8),
    pinlist: str | None = None,
    blank: str | list[str] | None = None,
    overlap: int = 102,
    rotation: int = 0,
    row_dist: float = 375 / 1.61,
    col_dist: float = 400 / 1.61,
    chip_type: str | None = None,
    min_button_diameter: int = 8,
    max_button_diameter: int = 30,
    chamber_diameter: int = 60,
    top_chamber: int | None = None,
    left_chamber: int | None = None,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.2,
    cluster_penalty: float = 50,
    roi_length: int | None = None,
    progress_bar: bool = False,
    search_timestep: int | list[int] = 0,
    search_channel: str | list[str] | None = None,
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> xr.Dataset | list[xr.Dataset]:
    """
    Find buttons in minichip images and standardize the resulting data in an xarray.Dataset

    Parameters
    ----------
    data :
        The input image data to be processed. It can be one of the following:
        - A file path or glob to image data.
        - An `xarray.DataArray` or `xarray.Dataset` containing image data.
        - A sequence of file paths, `xarray.DataArray`, or `xarray.Dataset`.
    shape :
        The shape of the button array, specifying the number of rows and columns in the
        image grid.
    pinlist :
        A file path to a CSV file that describes the tag to assign to each chamber on the
        chip. The CSV file must include a column called `Indices` that contains row and
        column pairs in the format `(row, col)`, and a `MutantID` column with the button
        names. Either `pinlist` or `shape` must be provided.
    blank :
        Values representing "blank" or non-expressed buttons in the pinlist, which will
        be replaced with empty strings in the dataset. Defaults to `["", "blank",
        "BLANK"]`.
    overlap :
        The number of pixels to exclude from the edges of adjacent tiles during the
        stitching process.
    rotation :
        The degree of rotation to apply to the stitched image.
    row_dist, col_dist :
        The distance between rows/columns of buttons in pixels.
    chip_type :
        The type of microfluidic chip that was imaged. Can be one of
        ["minichip"|"pc"|"ps"]. If `chip_type` is provided it overrides `row_dist` and
        `col_dist`.
    min_button_diameter, max_button_diameter :
        The minimum/maximum diameter in pixels for detecting buttons in the image.
    chamber_diameter :
        The diameter in pixels of the chamber around each button.
    top_chamber, left_chamber :
        The pixel offset of the edge of the top/leftmost chamber on the chip. If set to
        `None` the offset is automatically found.
    low_edge_quantile, high_edge_quantile :
        The lower/upper quantile for edge detection, used to identify the dimmest edges
        when detecting buttons. Must be between 0 and 1.
    num_iter :
        The maximum number of iterations to perform the button detection process using
        RANSAC. A higher value takes longer but finds buttons more accurately.
    min_roundness :
        The minimum roundness value for detected buttons. Buttons that do not meet this
        threshold are excluded. Must be between 0 and 1.
    cluster_penalty :
        A penalty number that balances two factors when identifying clusters: penalizing
        high inter-cluster variance and penalizing deviations from the expected number of
        items in a cluster. A higher value places more emphasis on the second factor.
    roi_length :
        The length in pixels of the region of interest (ROI) around detected buttons. If
        `None`, the ROI length is set to 1.2 * `chamber_diameter`.
    progress_bar :
        If True, display a progress bar during processing to track the progress of the
        pipeline.
    search_timestep :
        The timesteps on which to search for buttons. A timestep that is not in
        `search_timestep` uses button locations from the closest searched timestep before
        it, or if there isn't one, the closest timestep after it.
    search_channel :
        The channel or list of channels to use for button detection and expression
        filtering. If `None`, all channels will be used.
    roi_only :
        If True, only returns the region of interest from the dataset.
    drop_tiles :
        If True, removes the "tile" variable from the dataset after stitching.
    interactive:
        If True, open a window to visualize and tune image processing step-by-step.

    Returns
    -------
    Processed image(s) : xr.Dataset | list[xr.Dataset]
        The processed image in xr.Dataset from the executed pipeline as the outcome of
        the image pipeline workflow.

    Notes
    -----
    This function uses a pipeline architecture to process image data. The steps in the
    pipeline consist of:
    - 'read' : Reads in an input set of zarr, tiff, ome-tiff, or xarray and associated
      metadata.
    - 'standardize_format' : Transforms the dataset into a consistent format for
      downstream processing.
    - 'identify_buttons' : Identifies buttons based on the provided `pinlist` or `shape`
      parameters.
    - 'stitch' : Stitches image tiles based on the overlap parameter.
    - 'find_buttons' : Detects buttons based on edge detection and clustering.
    - 'drop' : Optionally removes elements in the dataset based on the `roi_only` and
      `drop_tiles` options.
    - 'restore_format' : Transforms the dataset back into the original format it was
      read in as.

    Examples
    --------
    >>> processed_mini_chip = mini_chip(
    ...     data=my_image_data, channels=[0], pinlist="pinlist.csv", overlap=100, rotation=45
    ... )

    This processes `my_image_data` by stitching tiles with 100 pixels of overlap,
    rotating the image by 45 degrees, using channel 0 for detection, and using the
    chamber layout from 'pinlist.csv'.
    """
    pipe = microfluidic_chip_pipe(
        shape=shape,
        pinlist=pinlist,
        blank=blank,
        overlap=overlap,
        rotation=rotation,
        row_dist=row_dist,
        col_dist=col_dist,
        chip_type=chip_type,
        min_button_diameter=min_button_diameter,
        max_button_diameter=max_button_diameter,
        chamber_diameter=chamber_diameter,
        top_chamber=top_chamber,
        left_chamber=left_chamber,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        cluster_penalty=cluster_penalty,
        roi_length=roi_length,
        progress_bar=progress_bar,
        search_timestep=search_timestep,
        search_channel=search_channel,
        roi_only=roi_only,
        drop_tiles=drop_tiles,
        interactive=interactive,
    )
    return pipe(data=data)


def microfluidic_chip_pipe(
    shape: tuple[int, int] = (8, 8),
    pinlist: str | None = None,
    blank: str | list[str] | None = None,
    overlap: int = 102,
    rotation: int = 0,
    row_dist: float = 375 / 1.61,
    col_dist: float = 400 / 1.61,
    chip_type: str | None = None,
    min_button_diameter: int = 8,
    max_button_diameter: int = 30,
    chamber_diameter: int = 60,
    top_chamber: int | None = None,
    left_chamber: int | None = None,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.2,
    cluster_penalty: float = 50,
    roi_length: int | None = None,
    progress_bar: bool = False,
    search_timestep: int | list[int] = 0,
    search_channel: str | list[str] | None = None,
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> Pipeline:
    """
    Build a Pipeline object that detects buttons in images of microfluidic chips and
    standardizes the resulting data in an xarray.Dataset.

    Reference
    -------
    For detailed information on how to use this pipeline, refer to :func:`microfluidic_chip`.
    """
    if chip_type is not None:
        if chip_type == "minichip":
            row_dist, col_dist = 375 / 1.61, 400 / 1.61
        elif chip_type == "pc":
            row_dist, col_dist = 406 / 3.22, 750 / 3.22
        elif chip_type == "ps":
            row_dist, col_dist = 375 / 3.22, 655 / 3.22
        else:
            raise ValueError(
                f"Invalid chip type: {chip_type}. Must be one of ['pc', 'ps', 'minichip']"
            )

    pipe = Pipeline("read")
    pipe.add_pipe("standardize_format")
    pipe.add_pipe("identify_buttons", shape=shape, pinlist=pinlist, blank=blank)
    pipe.add_pipe("stitch", overlap=overlap)
    pipe.add_pipe("rotate", rotation=rotation)
    pipe.add_pipe(
        "find_buttons",
        row_dist=row_dist,
        col_dist=col_dist,
        min_button_diameter=min_button_diameter,
        max_button_diameter=max_button_diameter,
        chamber_diameter=chamber_diameter,
        top_chamber=top_chamber,
        left_chamber=left_chamber,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        cluster_penalty=cluster_penalty,
        roi_length=roi_length,
        progress_bar=progress_bar,
        search_timestep=search_timestep,
        search_channel=search_channel,
        interactive=interactive,
    )
    pipe.add_pipe("drop", roi_only=roi_only, drop_tiles=drop_tiles)
    pipe.add_pipe("restore_format")

    return pipe


def mrbles(
    data: ArrayLike | str,
    spectra: str | PathLike | StringIO,
    codes: str | PathLike | StringIO,
    flatfield: float = 1.0,
    darkfield: float = 0.0,
    overlap: int = 102,
    min_bead_diameter: int = 10,
    max_bead_diameter: int = 50,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.3,
    roi_length: int | None = None,
    search_channel: str | list[str] | None = None,
    reference: str = "eu",
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> xr.Dataset | list[xr.Dataset]:
    """
    Find MRBLEs (Microspheres with Ratiometric Barcode Lanthanide Encoding Beads) in
    images and standardize the resulting data in an xarray.Dataset.

    Parameters
    ----------
    data :
        The input image data to be processed. It can be one of the following:
        - A file path or glob to image data.
        - An `xarray.DataArray` or `xarray.Dataset` containing image data.
        - A sequence of file paths, `xarray.DataArray`, or `xarray.Dataset`.
    spectra :
        A path to a CSV file that contains the reference spectrum of each lanthanide
        used for decoding.
    codes :
        A path to a CSV file that contains the name and lanthanide content of each code.
    flatfield :
        The flatfield correction factor or path to a flatfield correction image.
    darkfield :
        The darkfield correction factor or path to a darkfield correction image.
    overlap :
        The number of pixels to exclude from the edges of adjacent tiles during the
        stitching process.
    min_bead_diameter, max_bead_diameter :
        The minimum/maximum diameter in pixels for detecting beads in the image.
    low_edge_quantile, high_edge_quantile :
        The lower/upper quantile for edge detection, tuning how dim edges can be when
        detecting beads. Must be between 0 and 1.
    num_iter :
        The maximum number of iterations to perform the bead detection process using
        RANSAC. A higher value will take longer but will find beads more accurately.
    min_roundness :
        The minimum roundness value for detected beads. Beads below this threshold are
        excluded. Must be between 0 and 1.
    roi_length :
        The length in pixels of the region of interest (ROI) around detected beads. If
        None, the ROI length is set to 2 * `max_bead_diameter`.
    search_channel :
        The channel or list of channels to use for bead detection. If `None`, all
        channels will be used.
    reference :
        The reference material or standard used for spectral decoding. The default is
        "eu" (Europium), which is typically used in MRBLEs for spectral comparison.
    roi_only :
        If True, only returns the region of interest from the dataset, ignoring other
        parts of the image.
    drop_tiles :
        If True, removes the "tile" variable from the dataset after stitching.
    interactive:
        If True, open a window to visualize and tune image processing step-by-step.

    Returns
    -------
    Processed image(s) : xr.Dataset | list[xr.Dataset]
        The processed images and ROI.

    Notes
    -----
    This function uses a pipeline architecture to process MRBLEs image data. The steps
    in the pipeline include:
    - 'read' : Reads in an input set of zarr, tiff, ome-tiff, or xarray and associated
      metadata.
    - 'standardize_format' : Transforms the dataset into a consistent format for
      downstream processing.
    - 'flatfield_correct' : Applies flatfield and darkfield corrections to the image
      data.
    - 'stitch' : Stitches image tiles based on the overlap parameter.
    - 'find_beads' : Detects beads based on specified diameter, roundness, and other
      parameters.
    - 'identify_mrbles' : Assign a code to each bead by matching spectral signatures to
      the provided reference spectra.
    - 'drop' : Optionally removes elements in the dataset based on the `roi_only` and
      `drop_tiles` options.
    - 'restore_format' : Transforms the dataset back into the original format it was
      read in as.

    Examples
    --------
    >>> processed_mrbles = mrbles(
    ...     spectra=my_spectra, codes=my_codes, data=my_image_data, channels=[0], overlap=100
    ... )

    This processes `my_image_data` by stitching tiles with 100 pixels of overlap, using
    channel 0 for analysis, and matching bead spectral signatures against `my_spectra`
    and `my_codes`.
    """
    pipe = mrbles_pipe(
        spectra=spectra,
        codes=codes,
        flatfield=flatfield,
        darkfield=darkfield,
        overlap=overlap,
        min_bead_diameter=min_bead_diameter,
        max_bead_diameter=max_bead_diameter,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        roi_length=roi_length,
        search_channel=search_channel,
        reference=reference,
        roi_only=roi_only,
        drop_tiles=drop_tiles,
        interactive=interactive,
    )
    return pipe(data=data)


def mrbles_pipe(
    spectra: str | PathLike | StringIO,
    codes: str | PathLike | StringIO,
    flatfield: float = 1.0,
    darkfield: float = 0.0,
    overlap: int = 102,
    min_bead_diameter: int = 10,
    max_bead_diameter: int = 50,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.3,
    roi_length: int | None = None,
    search_channel: str | list[str] | None = None,
    reference: str = "eu",
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> Pipeline:
    """
    Build a Pipeline object that can detect MRBLEs (Microspheres with Ratiometric
    Barcode Lanthanide Encoding Beads) in images and standardize the resulting data in
    an xarray.Dataset.

    Reference
    -------
    This function builds the necessary pipeline for detecting beads in MRBLE images.
    For detailed information on how to use this pipeline, refer to :func:`mrbles`.
    """
    pipe = Pipeline("read")
    pipe.add_pipe("standardize_format")
    pipe.add_pipe("flatfield_correct", flatfield=flatfield, darkfield=darkfield)
    pipe.add_pipe("stitch", overlap=overlap)
    pipe.add_pipe(
        "find_beads",
        min_bead_diameter=min_bead_diameter,
        max_bead_diameter=max_bead_diameter,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        roi_length=roi_length,
        search_channel=search_channel,
        interactive=interactive,
    )
    pipe.add_pipe("identify_mrbles", spectra=spectra, codes=codes, reference=reference)
    pipe.add_pipe("drop", roi_only=roi_only, drop_tiles=drop_tiles)
    pipe.add_pipe("restore_format")

    return pipe


def beads(
    data: ArrayLike | str,
    flatfield: float = 1.0,
    darkfield: float = 0.0,
    overlap: int = 102,
    min_bead_diameter: int = 10,
    max_bead_diameter: int = 50,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.3,
    roi_length: int | None = None,
    search_channel: str | list[str] | None = None,
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> xr.Dataset | list[xr.Dataset]:
    """
    Find beads in images and standardize the resulting data in an xarray.Dataset.

    Parameters
    ----------
    data :
        The input image data to be processed. It can be one of the following:
        - A file path or glob to image data.
        - An `xarray.DataArray` or `xarray.Dataset` containing image data.
        - A sequence of file paths, `xarray.DataArray`, or `xarray.Dataset`.
    flatfield :
        The flatfield correction factor or path to a flatfield correction image.
    darkfield :
        The darkfield correction factor or path to a darkfield correction image.
    overlap :
        The number of pixels to exclude from the edges of adjacent tiles during the
        stitching process.
    min_bead_diameter, max_bead_diameter :
        The minimum/maximum diameter in pixels for detecting beads in the image.
    low_edge_quantile, high_edge_quantile :
        The lower/upper quantile for edge detection, tuning how dim edges can be when
        detecting beads. Must be between 0 and 1.
    num_iter :
        The maximum number of iterations to perform the bead detection process using
        RANSAC. A higher value will take longer but will find beads more accurately.
    min_roundness :
        The minimum roundness value for detected beads. Beads that do not meet this
        threshold are excluded. Must be between 0 and 1.
    roi_length :
        The length in pixels of the region of interest (ROI) around detected beads. If
        None, the ROI length is set to 2 * `max_bead_diameter`.
    search_channel :
        The channel or list of channels to use for bead detection. If `None`, all
        channels will be used.
    roi_only :
        If True, only returns the region of interest from the dataset.
    drop_tiles :
        If True, removes the "tile" variable from the dataset after stitching.
    interactive:
        If True, open a window to visualize and tune image processing step-by-step.

    Returns
    -------
    Processed image(s) : xr.Dataset | list[xr.Dataset]
        The processed images and ROI.

    Notes
    -----
    This function uses a pipeline architecture to process image data. The steps in the
    pipeline include:

    - 'read' : Reads in an input set of zarr, tiff, ome-tiff, or xarray and associated
      metadata.
    - 'standardize_format' : Transforms the dataset into a consistent format for
      downstream processing.
    - 'flatfield_correct' : Applies flatfield and darkfield corrections to the image
      data.
    - 'stitch' : Stitches image tiles based on the overlap parameter.
    - 'find_beads' : Detects beads based on specified diameter, roundness, and other
      parameters.
    - 'drop' : Optionally removes elements in the dataset based on the `roi_only` and
      `drop_tiles` options.
    - 'restore_format' : Transforms the dataset back into the original format it was
      read in as.

    Examples
    --------
    >>> bead_image = beads(
    ...     data=my_image_data,
    ...     channels=[0, 1],
    ...     overlap=100,
    ...     min_bead_diameter=10,
    ...     max_bead_diameter=40,
    ... )

    This processes `my_image_data` by stitching tiles with 100 pixels of overlap, using
    only channels 0 and 1, and detecting beads with a diameter between 10 and 40 pixels.
    """
    pipe = beads_pipe(
        flatfield=flatfield,
        darkfield=darkfield,
        overlap=overlap,
        min_bead_diameter=min_bead_diameter,
        max_bead_diameter=max_bead_diameter,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        roi_length=roi_length,
        search_channel=search_channel,
        roi_only=roi_only,
        drop_tiles=drop_tiles,
        interactive=interactive,
    )
    return pipe(data=data)


def beads_pipe(
    flatfield: float = 1.0,
    darkfield: float = 0.0,
    overlap: int = 102,
    min_bead_diameter: int = 5,
    max_bead_diameter: int = 25,
    low_edge_quantile: float = 0.1,
    high_edge_quantile: float = 0.9,
    num_iter: int = 5000000,
    min_roundness: float = 0.3,
    roi_length: int | None = None,
    search_channel: str | list[str] | None = None,
    roi_only: bool = False,
    drop_tiles: bool = True,
    interactive: bool = False,
) -> Pipeline:
    """
    Build a Pipeline object that can detect beads in images and standardize the
    resulting data in an xarray.Dataset.

    Reference
    -------
    This function builds the necessary pipeline for detecting beads.
    For detailed information on how to use this pipeline, refer to :func:`beads`.
    """
    pipe = Pipeline("read")
    pipe.add_pipe("standardize_format")
    pipe.add_pipe("flatfield_correct", flatfield=flatfield, darkfield=darkfield)
    pipe.add_pipe("stitch", overlap=overlap)
    pipe.add_pipe(
        "find_beads",
        min_bead_diameter=min_bead_diameter,
        max_bead_diameter=max_bead_diameter,
        low_edge_quantile=low_edge_quantile,
        high_edge_quantile=high_edge_quantile,
        num_iter=num_iter,
        min_roundness=min_roundness,
        roi_length=roi_length,
        search_channel=search_channel,
        interactive=interactive,
    )
    pipe.add_pipe("drop", roi_only=roi_only, drop_tiles=drop_tiles)
    pipe.add_pipe("restore_format")

    return pipe


def image(
    data: ArrayLike | str,
    overlap: int = 102,
    rotation: float = 0,
    roi_only: bool = False,
    drop_tiles: bool = True,
) -> xr.Dataset | list[xr.Dataset]:
    """
    Read in images and standardize the resulting data in an xarray.Dataset.

    Parameters
    ----------
    data :
        The input image data to be processed. It can be one of the following:
        - A file path or glob to image data.
        - An `xarray.DataArray` or `xarray.Dataset` containing image data.
        - A sequence of file paths, `xarray.DataArray`, or `xarray.Dataset`.
    overlap :
        The number of pixels to exclude from the edges of adjacent tiles during the
        stitching process.
    roi_only :
        If True, only returns the region of interest from the dataset.
    drop_tiles :
        If True, removes the "tile" variable from the dataset after stitching.
    Returns
    -------
    Processed image(s) : xr.Dataset | list[xr.Dataset]
        The processed images and ROI.

    Notes
    -----
    This function uses a pipeline architecture to process image data. The steps in the
    pipeline include:

    - 'read' : Reads in an input set of zarr, tiff, ome-tiff, or xarray and associated
      metadata.
    - 'standardize_format' : Transforms the dataset into a consistent format for
      downstream processing.
    - 'stitch' : Stitches image tiles based on the overlap parameter.
    - 'drop' : Optionally removes elements in the dataset based on the `roi_only` and
      `drop_tiles` options.
    - 'restore_format' : Transforms the dataset back into the original format it was
      read in as.

    Examples
    --------
    >>> processed_image = image(data=my_image_data, channels=[0, 1], overlap=100, rotation=45)
    """
    pipe = image_pipe(
        overlap=overlap,
        rotation=rotation,
        roi_only=roi_only,
        drop_tiles=drop_tiles,
    )
    return pipe(data=data)


def image_pipe(
    overlap: int = 102,
    rotation: float = 0,
    roi_only: bool = False,
    drop_tiles: bool = True,
) -> Pipeline:
    """
    Build a Pipeline object that reads in an image and standardizes the resulting data
    in an xarray.Dataset.

    Reference
    -------
    This function builds the necessary pipeline for customizing an image-processing pipeline.
    For detailed information on how to use this pipeline, refer to :func:`image`.
    """
    pipe = Pipeline("read")
    pipe.add_pipe("standardize_format")
    pipe.add_pipe("stitch", overlap=overlap)
    pipe.add_pipe("rotate", rotation=rotation)
    pipe.add_pipe("drop", roi_only=roi_only, drop_tiles=drop_tiles)
    pipe.add_pipe("restore_format")
    return pipe



import os
import pathlib

import dask.array as da
import dask_image.ndinterp
import tifffile
import xarray as xr

from magnify import registry, utils


@registry.component("standardize_format")
def standardize_format(xp: xr.Dataset | xr.DataArray) -> xr.Dataset:
    if isinstance(xp, xr.DataArray):
        xp = xr.Dataset({"tile": xp}).assign_attrs(xp.attrs)

    # Rename dimensions since we'll be adding new arrays whose names will conflict.
    for old_name in ["x", "y", "row", "col"]:
        if old_name in xp.tile.dims:
            xp = xp.rename({old_name: "tile_" + old_name})

    # Save the original dimension ordering to be restored later.
    xp.attrs["__original_tile_dims__"] = list(xp.tile.dims)

    desired_order = ["channel", "time", "tile_row", "tile_col", "tile_y", "tile_x"]
    # If we have additional dimensions stack them all into a single time dimension.
    extra_dims = [dim for dim in xp.tile.dims if dim not in desired_order]
    if len(extra_dims) > 0:
        if "time" in xp.tile.dims:
            # Rename the time dimension to avoid conflicts.
            xp = xp.rename(time="__time__")
            extra_dims.append("__time__")
        xp = xp.stack(time=extra_dims)

    # Reorder the dimensions so they're always consistent and add missing dimensions.
    for dim in desired_order:
        if dim not in xp.tile.dims:
            xp["tile"] = xp.tile.expand_dims(dim)

    xp = xp.transpose(*desired_order)

    return xp


@registry.component("rename_labels")
def rename_labels(xp: xr.Dataset, **coords):
    for coord_name, new_labels in coords.items():
        if isinstance(new_labels, dict):
            xp = xp.assign_coords({coord_name: xp[coord_name].to_series().replace(new_labels)})
        else:
            xp = xp.assign_coords({coord_name: new_labels})
    return xp


@registry.component("rotate")
def rotate(xp: xr.Dataset, rotation=0):
    xp["image"].data = dask_image.ndinterp.rotate(
        xp.image.data, rotation, axes=(-1, -2), reshape=False
    )
    return xp


@registry.component("flatfield_correct")
def flatfield_correct(xp: xr.Dataset, flatfield=1.0, darkfield=0.0):
    if isinstance(flatfield, os.PathLike | str):
        flatfield = pathlib.Path(flatfield).expanduser()
        if flatfield.is_dir():
            # We have a zarr directory with both flatfield and darkfield information.
            flatfield = xr.dataset.open_zarr(flatfield, group="flatfield")["flatfield"]
            # Account for channel mismatches.
            flatfield = xr.align(xp.tile, flatfield, join="left")[1].fillna(
                flatfield.sel(channel="default")
            )
            darkfield = xr.dataset.open_zarr(flatfield, group="darkfield")["darkfield"]
        else:
            with tifffile.TiffFile(flatfield) as tif:
                flatfield = tif.asarray()

    if isinstance(darkfield, os.PathLike | str):
        darkfield = pathlib.Path(darkfield).expanduser()
        with tifffile.TiffFile(darkfield) as tif:
            darkfield = tif.asarray()

    tiles = (xp.tile.astype(float) - darkfield).clip(min=0)
    max_val = tiles.max()
    tiles = tiles / flatfield
    tiles = tiles * max_val / tiles.max()
    xp["tile"] = tiles.astype(xp.tile.dtype)
    return xp


@registry.component("basic_correct")
def basic_correct(xp: xr.Dataset):
    import basicpy

    # Iterate over one channel at a time to avoid memory issues.
    for channel in xp.channel:
        tiles = xp.tile.sel(channel=channel)
        # Only fit to the first timestep since variation in time won't help with the correction.
        train_tiles = tiles.isel(time=0).to_numpy().reshape(-1, tiles.shape[-2], tiles.shape[-1])
        model = basicpy.basicpy.BaSiC(get_darkfield=True, smoothness_flatfield=1)
        model.fit(train_tiles)

        # Apply the flatfield correction.
        def transform(block, model=model):
            init_shape = block.shape
            # Reshape to handle cases where each block isn't just a single tile.
            block = block.reshape(-1, block.shape[-2], block.shape[-1])
            block = model.transform(block)
            return block.reshape(init_shape)

        tiles = da.map_blocks(transform, tiles.data, dtype=tiles.dtype)
        xp["tile"].loc[{"channel": channel}] = tiles

    xp.mg.cache("tile")
    return xp


@registry.component("horizontal_flip")
def horizontal_flip(xp: xr.Dataset):
    if "image" in xp:
        xp["image"] = xp.image.isel(im_x=slice(None, None, -1))
    else:
        xp["tile"] = xp.tile.isel(tile_x=slice(None, None, -1))
    return xp


@registry.component("vertical_flip")
def vertical_flip(xp: xr.Dataset):
    if "image" in xp:
        xp["image"] = xp.image.isel(im_y=slice(None, None, -1))
    else:
        xp["tile"] = xp.tile.isel(tile_y=slice(None, None, -1))
    return xp


@registry.component("circle_mask")
def circle_mask(
    xp: xr.Dataset,
    center: tuple[int, int],
    diameter: int,
    mask_inner=False,
):
    radius = diameter // 2
    img_shape = (xp.image.shape if "image" in xp else xp.tile.shape)[-2:]

    mask = utils.circle(img_shape, center, radius, True)
    mask = ~mask if mask_inner else mask
    if "image" in xp:
        xp["image"] *= mask
    else:
        xp["tile"] *= mask

    return xp



import tempfile

import dask.array as da
import xarray as xr

from magnify import utils

cache = []


@xr.register_dataset_accessor("mg")
class MagnifyAccessor:
    def __init__(self, dataset):
        self._dataset = dataset
        self._tempdir = tempfile.TemporaryDirectory()
        cache.append(self._tempdir)

    def cache(self, variables=None):
        variables = utils.to_list(variables)
        if variables:
            arrays = {name: self._dataset.variables[name] for name in variables}
        else:
            arrays = self._dataset.variables

        for name, arr in arrays.items():
            if isinstance(arr.data, da.Array):
                arr.data = da.to_zarr(
                    arr.data,
                    url=self._tempdir.name,
                    component=name,
                    return_stored=True,
                    overwrite=True,
                )

        return self._dataset



from collections.abc import Callable

import xarray as xr
from numpy.typing import ArrayLike

import magnify.registry as registry


class Pipeline:
    def __init__(self, reader: str):
        self.reader: Callable[[ArrayLike | str], xr.Dataset] = registry.readers.get(reader)()
        self.components: list[tuple[str, Callable[[xr.Dataset], xr.Dataset]]] = []

    def __call__(
        self,
        data: ArrayLike | str,
    ) -> xr.Dataset | list[xr.Dataset]:
        inputs = self.reader(data=data)
        assays = []
        for assay in inputs:
            for _, component in self.components:
                assay = component(assay)

            assays.append(assay)

        if len(assays) == 1:
            assays = assays[0]

        return assays

    def add_pipe(
        self,
        component: str | Callable[..., xr.Dataset],
        name: str | None = None,
        after: str | int | None = None,
        before: str | int | None = None,
        first: bool = False,
        last: bool = False,
        **kwargs,
    ) -> None:
        if isinstance(component, str):
            if name is None:
                name = component
            component_factory = registry.components.get(component)
            func = component_factory(**kwargs)
        else:
            name = component.__name__ if name is None else name

            def func(xp):
                return component(xp, **kwargs)

        if after is None and before is None and not first and not last:
            last = True
        if (after is not None) + (before is not None) + first + last > 1:
            raise ValueError("Only one of after, before, first, and last can be set.")

        # Check that the new component's name is unique
        if self.components and name in next(zip(*self.components, strict=True)):
            raise ValueError(f"A component with the name '{name}' already exists in the pipeline.")

        # Find where to insert the component.
        new_element = (name, func)
        if first:
            idx = 0
        elif last:
            idx = len(self.components)
        elif isinstance(before, int):
            idx = before
        elif isinstance(before, str):
            idx = [name for (name, _) in self.components].index(before)
        elif isinstance(after, int):
            idx = after + 1
        elif isinstance(after, str):
            idx = [name for (name, _) in self.components].index(after) + 1
        else:
            raise ValueError("before/after must be a string or int.")

        self.components.insert(idx, new_element)

    def remove_pipe(self, name: str) -> None:
        if not self.components:
            raise ValueError(f"Cannot remove pipe '{name}': pipeline has no components")
        component_names = list(zip(*self.components, strict=True))[0]
        if name not in component_names:
            raise ValueError(f"Component '{name}' not found in pipeline")
        idx = component_names.index(name)
        self.components.pop(idx)
